package edu.stanford.db.rdf.model.i;

import org.w3c.rdf.vocabulary.rdf_syntax_19990222.RDF;
import org.w3c.rdf.model.*;
import java.util.*;

/**
 * We can have two equal RDFNodeImpl with two different IDs
 */

public class NodeFactoryImpl implements NodeFactory {

  // string -> ResourceImpl
  Map rmap = new WeakHashMap();
  // string -> LiteralImpl
  Map lmap = new WeakHashMap();

  int lastNodeID;
  // TODO: be more robust
  static String uniquePrefix = "uri:rdf:" + System.currentTimeMillis() + "-";

  public NodeFactoryImpl() {
  }

  public synchronized int getLastNodeID() {

    return lastNodeID;
  }

  public Collection getResources() {

    return rmap.values();
  }

  public Collection getLiterals() {

    return lmap.values();
  }

  /**
   * Creates a resource out of a string
   */
  public synchronized Resource createResource(String str) {

    // check whether we have it in the registry
    Resource r = (Resource)rmap.get(str);
    if(r == null) {
      r = new ResourceImpl(getUnusedNodeID(), str);
      rmap.put(str, r);
    }
    return r;
  }

  /**
   * Creates a literal out of a string
   */
  public Literal createLiteral(String str) {

    // check whether we have it in the registry
    Literal r = (Literal)lmap.get(str);
    if(r == null) {
      r = new LiteralImpl(getUnusedNodeID(), str);
      lmap.put(str, r);
    }
    return r;
  }

  public int getUnusedNodeID() {

    return lastNodeID++;
  }

  /**
   * Creates a triple
   */
  public Statement createStatement(Resource subject, Resource predicate, RDFNode object) throws ModelException {
		
    // remap resources if necessary
    if(!(subject instanceof ResourceImpl))
      subject = createResource(subject.getLabel());
    if(!(predicate instanceof ResourceImpl))
      predicate = createResource(predicate.getLabel());
    if(object instanceof Resource) {
      if(!(predicate instanceof ResourceImpl))
	object = createResource(object.getLabel());
    } else if(object instanceof Literal) {
      if(!(object instanceof LiteralImpl))
	object = createLiteral(object.getLabel());
    }

    return (subject != null && predicate != null && object != null) ?
      new StatementImpl(subject, predicate, object) : null;
  }

  /**
   * Creates a resource with a unique ID
   */
  public synchronized Resource createUniqueResource() {

    // TODO: don't give it a name for now
    int nodeID = getUnusedNodeID();
    ResourceImpl node = new ResourceImpl(nodeID, uniquePrefix + nodeID);
    return node;
  }

  /**
   * Creates an ordinal property (rdf:li, rdf:_N)
   */
  public Resource createOrdinal(int i) {
    
    if(i < 1)
      throw new IllegalArgumentException("Attempt to construct invalid ordinal resource");
    else
      return createResource(RDF._Namespace + "_" + i);
  }

  /**
   * Creates a literal from a boolean
   */
  public Literal createLiteral(boolean b) throws ModelException {

    return createLiteral(String.valueOf(b));
  }

  /**
   * Creates a literal from a byte
   */
  public Literal createLiteral(byte b) throws ModelException {

    return createLiteral(String.valueOf(b));
  }


  /**
   * Creates a literal from a char
   */
  public Literal createLiteral(char b) throws ModelException {

    return createLiteral(String.valueOf(b));
  }


  /**
   * Creates a literal from a short
   */
  public Literal createLiteral(short b) throws ModelException {

    return createLiteral(String.valueOf(b));
  }


  /**
   * Creates a literal from an int
   */
  public Literal createLiteral(int b) throws ModelException {

    return createLiteral(String.valueOf(b));
  }


  /**
   * Creates a literal from a long
   */
  public Literal createLiteral(long b) throws ModelException{

    return createLiteral(String.valueOf(b));
  }


  /**
   * Creates a literal from a float
   */
  public Literal createLiteral(float b) throws ModelException{

    return createLiteral(String.valueOf(b));
  }


  /**
   * Creates a literal from a double
   */
  public Literal createLiteral(double b) throws ModelException{

    return createLiteral(String.valueOf(b));
  }



}
